/*This file is prepared for Doxygen automatic documentation generation.*/
/*! \file ******************************************************************
 *
 * \brief Main file of the USB MS ISP.
 *
 * - Compiler:           IAR EWAVR32 and GNU GCC for AVR32
 * - Supported devices:  All AVR32 devices with a USB module can be used.
 * - AppNote:
 *
 * \author               Atmel Corporation: http://www.atmel.com \n
 *                       Support and FAQ: http://support.atmel.no/
 *
 ***************************************************************************/

/* Copyright (C) 2006-2008, Atmel Corporation All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * 3. The name of ATMEL may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY ATMEL ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE EXPRESSLY AND
 * SPECIFICALLY DISCLAIMED. IN NO EVENT SHALL ATMEL BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


//_____  I N C L U D E S ___________________________________________________

#include <avr32/io.h>
#include <string.h>
#include "preprocessor.h"
#include "board.h"
#include "pm.h"
#include "flashc.h"

// all includes, defines and code related to USB commented
// #include "conf_usb.h"
// #include "usb_task.h"
// #if USB_HOST_FEATURE == ENABLED
// #include "host_mass_storage_task.h"
// #endif

#include "fs_com.h"
#include "file.h"
#include "navigation.h"
#include "crc32.h"
#include "conf_isp.h"

// extra includes
#include "gpio.h"
#include "usart.h"
#include "spi.h"
#include "sd_mmc_spi.h"
#include "lcd_hd44780.h"


//_____ M A C R O S ________________________________________________________


//_____ D E F I N I T I O N S ______________________________________________

#define FGCLK_USBB                48000000

#define FOSC                      ATPASTE2(FOSC, ISP_OSC)
#define OSC_STARTUP               ATPASTE3(OSC, ISP_OSC, _STARTUP)

#define pm_enable_osc_crystal     ATPASTE3(pm_enable_osc, ISP_OSC, _crystal)
#define pm_enable_clk             ATPASTE2(pm_enable_clk, ISP_OSC)

// all includes, defines and code related to USB commented
//#if USB_HOST_FEATURE == ENABLED
static crc32_table_t crc32_table;
//#endif


/*!
 *  Start the generation of system clocks
 */
static void sys_clk_gen_start(void)
{
#if FOSC !=  8000000 &&                                                        \
    FOSC != 12000000 &&                                                        \
    FOSC != 16000000
  #error Unsupported oscillator frequency
#endif

  volatile avr32_pm_t *const pm = &AVR32_PM;

  // Start the oscillator
  pm_enable_osc_crystal(pm, FOSC);
  pm_enable_clk(pm, OSC_STARTUP);

  // Set PLL0 VCO @ 96 MHz
  pm_pll_setup(pm, 0,                     // pll
                   FGCLK_USBB / FOSC - 1, // mul
                   0,                     // div
                   ISP_OSC,               // osc
                   63);                   // lockcount

  // Set PLL0 @ 48 MHz
  pm_pll_set_option(pm, 0,  // pll
                        1,  // pll_freq
                        1,  // pll_div2
                        0); // pll_wbwdisable

  // Enable PLL0
  pm_pll_enable(pm, 0);

  // Wait for PLL0 locked
  pm_wait_for_pll0_locked(pm);

  // fPBA: 12 MHz
  // fPBB: 12 MHz
  // fHSB: 12 MHz
  pm_cksel(pm, 1,   // pbadiv
               1,   // pbasel
               1,   // pbbdiv
               1,   // pbbsel
               1,   // hsbdiv
               1);  // hsbsel

  // Use 0 flash wait state
  flashc_set_wait_state(0);

  // Switch the main clock to PLL0
  pm_switch_to_clock(pm, AVR32_PM_MCCTRL_MCSEL_PLL0);

  // Setup USB GCLK
  pm_gc_setup(pm, AVR32_PM_GCLK_USBB, // gc
                  1,                  // osc_or_pll: use Osc (if 0) or PLL (if 1)
                  0,                  // pll_osc: select Osc0/PLL0 or Osc1/PLL1
                  0,                  // diven
                  0);                 // div

  // Enable USB GCLK
  pm_gc_enable(pm, AVR32_PM_GCLK_USBB);
}



// Synth Core  23-11-2011  Tolaemon
// Function: void FILE_SYS_SD_MMC_resources_init()
// Function which prepares the SPI to be used with
// the MMC_SD card
// Receives:
// Returns:
// Note:
//     Call before starting the OS scheduler
void FILE_SYS_SD_MMC_resources_init(void)
{

	// GPIO map for SPI.
	static const gpio_map_t SD_SPI_GPIO_MAP =
	{
		{ SD_MMC_SPI_SCK_PIN, SD_MMC_SPI_SCK_FUNCTION },  // SPI Clock.
		{ SD_MMC_SPI_MISO_PIN, SD_MMC_SPI_MISO_FUNCTION },// MISO.
		{ SD_MMC_SPI_MOSI_PIN, SD_MMC_SPI_MOSI_FUNCTION}, // MOSI.
		{ SD_MMC_SPI_NPCS_PIN, SD_MMC_SPI_NPCS_FUNCTION } // Chip Select NPCS.
	};

	// Options for SPI.
	spi_options_t spiOptions =
	{
		.reg          = SD_MMC_SPI_NPCS,
		.baudrate     = 1200000,//SD_MMC_SPI_MASTER_SPEED,
		.bits         = 8,
		.spck_delay   = 0,
		.trans_delay  = 0,
		.stay_act     = 1,
		.spi_mode     = 0,
		.modfdis      = 1
	};

	// Assign I/Os to SPI.
	gpio_enable_module(SD_SPI_GPIO_MAP,
	sizeof(SD_SPI_GPIO_MAP) / sizeof(SD_SPI_GPIO_MAP[0]));

	// Initialize as master.
	spi_initMaster(SD_MMC_SPI, &spiOptions);

	// Set selection mode: variable_ps, pcs_decode, delay.
	spi_selectionMode(SD_MMC_SPI, 0, 0, 0);

	// Enable SPI.1
	spi_enable(SD_MMC_SPI);

	sd_mmc_spi_init(spiOptions,FOSC0);

}//SD_MMC_resources_init



// #if USB_HOST_FEATURE == ENABLED

static void isp_task(void)
{
	  #define PROGRAM_START_PAGE  (PROGRAM_START_OFFSET / AVR32_FLASHC_PAGE_SIZE)
	  #define HEADER_SIZE         (sizeof(ID) + sizeof(UUID) + sizeof(crc32))

	  volatile avr32_usart_t  *usart;

	  static const U8 ID[] = ISP_FW_UPGRADE_FILE_ID;
	  static const U8 UUID[] = ISP_FW_UPGRADE_FILE_UUID;

	  static enum
	  {
		ISP_STATE_ERROR               = -1,
		ISP_STATE_IDLE                =  0,
		ISP_STATE_MOUNT               =  1,
		ISP_STATE_CHECK               =  2,
		ISP_STATE_CRC32               =  3,
		ISP_STATE_ERASE_AND_VERIFY    =  4,
		ISP_STATE_PROGRAM_AND_VERIFY  =  5
	  } isp_state = ISP_STATE_IDLE;

	  static U32 crc32;

	  #if __GNUC__ && __AVR32__
  		    __attribute__((__aligned__(4)))
	  #elif __ICCAVR32__
		    #pragma data_alignment = 4
	  #endif
	  
	  U8 buffer[AVR32_FLASHC_PAGE_SIZE];
	  volatile U64 *flash_ptr;
	  S32 i;

	  usart = serial_232_PORT_USART;

	  /* not using USB
	  if (!Is_host_ms_configured())
	  {
		isp_state = ISP_STATE_IDLE;
		return;
	  } */
  
	  switch (isp_state)
	  {
		  case ISP_STATE_ERROR:
  
			 // Clear the address AVR32_FLASHC_USER_PAGE_ADDRESS+ISP_FORCE_OFFSET ( 0x808001F8 )
			 // where the key that activates de ISP programming is stored
			 flash_ptr = (U64 *)AVR32_FLASHC_USER_PAGE;
			 memcpy(buffer, (void *)flash_ptr, AVR32_FLASHC_PAGE_SIZE);
			 *(U32 *)&buffer[ISP_FORCE_OFFSET] = 0xFFFFFFFF;
			 flashc_clear_page_buffer();
			 for (i = 0; i < AVR32_FLASHC_PAGE_SIZE / sizeof(*flash_ptr); i++)
			 *flash_ptr++ = ((U64 *)buffer)[i];
			 flashc_erase_user_page(FALSE);
			 flashc_write_user_page();
	 
			 // enables the watchdog to reset de system
			 Disable_global_interrupt();
			 AVR32_WDT.ctrl = AVR32_WDT_CTRL_EN_MASK |
			 (10 << AVR32_WDT_CTRL_PSEL_OFFSET) |
			 (AVR32_WDT_KEY_VALUE << AVR32_WDT_CTRL_KEY_OFFSET);
			 asm ("");
			 AVR32_WDT.ctrl = AVR32_WDT_CTRL_EN_MASK |
			 (10 << AVR32_WDT_CTRL_PSEL_OFFSET) |
			 ((~AVR32_WDT_KEY_VALUE << AVR32_WDT_CTRL_KEY_OFFSET) & AVR32_WDT_CTRL_KEY_MASK);
	 
			 // wati until the system resets
			 LCD_HD44780_write("ERROR.RESET!",2);
			 usart_write_line(usart, "ERROR.RESET!\r\n");
			 while (1){
				 usart_write_line(usart, ".");
			 };
  
			break;

		  case ISP_STATE_IDLE:
		  
			LCD_HD44780_write("MOUNT",2);
			usart_write_line(usart, "ISP_STATE_MOUNT\r\n");
			isp_state = ISP_STATE_MOUNT;
			break;

		  case ISP_STATE_MOUNT:
		  
			nav_reset();
			if (!nav_drive_set(0) || !nav_partition_mount()){
				
			  LCD_HD44780_write("ERR: MOUNT 1",2);
			  usart_write_line(usart, "ERR: ISP_STATE_MOUNT 1\r\n");
			  isp_state = ISP_STATE_ERROR;
			  break;			  
			  
			}//if			
			LCD_HD44780_write("CHECK",2);
			usart_write_line(usart, "ISP_STATE_CHECK\r\n");
			isp_state = ISP_STATE_CHECK;
			break;

		  case ISP_STATE_CHECK:
		  
			if (!nav_filelist_single_enable(FS_FILE) || !nav_filelist_set(0, FS_FIND_NEXT) || !nav_file_name(ISP_FW_UPGRADE_FILENAME, 0, FS_NAME_CHECK, TRUE)){
					
			   LCD_HD44780_write("ERR: CHECK 1",2);
			   usart_write_line(usart, "ERROR: ISP_STATE_CHECK 1\r\n");
			   isp_state = ISP_STATE_ERROR;
				  
			   break;
			}//if
	  
			i = nav_file_lgt() - HEADER_SIZE;
			if (i < 0 ||
				i > flashc_get_flash_size() - PROGRAM_START_OFFSET ||
				!file_open(FOPEN_MODE_R) ||
				file_read_buf(buffer, sizeof(ID)) != sizeof(ID) ||
				memcmp(buffer, ID, sizeof(ID)) ||
				file_read_buf(buffer, sizeof(UUID)) != sizeof(UUID) ||
				memcmp(buffer, UUID, sizeof(UUID)) ||
				file_read_buf((U8 *)&crc32, sizeof(crc32)) != sizeof(crc32))
			{
				  LCD_HD44780_write("ERR: CHECK 2",2);
				  usart_write_line(usart, "ERROR: ISP_STATE_CHECK 2\r\n");
				  isp_state = ISP_STATE_ERROR;
				  break;
			}

			crc32_reset();
			LCD_HD44780_write("CRC32",2);
			usart_write_line(usart, "ISP_STATE_CRC32\r\n");
			isp_state = ISP_STATE_CRC32;
			break;

		  case ISP_STATE_CRC32:

			switch (file_eof()){
				case 0x00:
					if (!(i = file_read_buf(buffer, sizeof(buffer))))
					{
					LCD_HD44780_write("ERR: CRC32 1",2);
					usart_write_line(usart, "ERROR: ISP_STATE_CRC32 1\r\n");
					isp_state = ISP_STATE_ERROR;
					break;
					}
					crc32_iterate_buffer(buffer, i);
					break;
				case 0x01:
					if (crc32 != crc32_get_value())
					{
					LCD_HD44780_write("ERR: CRC32 2",2);
					usart_write_line(usart, "ERROR: ISP_STATE_CRC32 2\r\n");
					isp_state = ISP_STATE_ERROR;
					break;
					}
					LCD_HD44780_write("ERASE_AND_VERIFY",2);
					usart_write_line(usart, "ISP_STATE_ERASE_AND_VERIFY\r\n");
					isp_state = ISP_STATE_ERASE_AND_VERIFY;
					break;
				case 0xFF:
					LCD_HD44780_write("ERR: CRC32 3",2);
					usart_write_line(usart, "ERROR: ISP_STATE_CRC32 3\r\n");
					isp_state = ISP_STATE_ERROR;
					break;
			}//switch (file_eof())
			break;

		case ISP_STATE_ERASE_AND_VERIFY:
		
			for (i = flashc_get_page_region(PROGRAM_START_PAGE); i < AVR32_FLASHC_REGIONS; i++)
			{
				flashc_lock_region(i, FALSE);
				if (flashc_is_region_locked(i))
				{
				LCD_HD44780_write("ERR: E_VERIFY 1",2);
				usart_write_line(usart, "ERROR: ISP_STATE_ERASE_AND_VERIFY 1\r\n");
				isp_state = ISP_STATE_ERROR;
				return;
				}
			}

			for (i = flashc_get_page_count() - 1; i >= PROGRAM_START_PAGE; i--)
			{
				if (!flashc_erase_page(i, TRUE))
				{
				LCD_HD44780_write("ERR: E_VERIFY 2",2);
				usart_write_line(usart, "ERROR: ISP_STATE_ERASE_AND_VERIFY 2\r\n");
				isp_state = ISP_STATE_ERROR;
				return;
				}
			}

			if (!file_seek(HEADER_SIZE, FS_SEEK_SET))
			{
				LCD_HD44780_write("ERR: E_VERIFY 3",2);
				usart_write_line(usart, "ERROR: ISP_STATE_ERASE_AND_VERIFY 3\r\n");
				isp_state = ISP_STATE_ERROR;
				break;
			}
			isp_state = ISP_STATE_PROGRAM_AND_VERIFY;
			break;

		  case ISP_STATE_PROGRAM_AND_VERIFY:
		  
			switch (file_eof()){
				
				case 0x00:
					if (!(i = file_getpos()))
					{
  					LCD_HD44780_write("ERR: P_VERIFY 1",2);
					usart_write_line(usart, "ERROR: ISP_STATE_PROGRAM_AND_VERIFY 1\r\n");
					isp_state = ISP_STATE_ERROR;
					break;
					}
					flash_ptr = (U64 *)(PROGRAM_START_ADDRESS + i - HEADER_SIZE);
					if (!(i = file_read_buf(buffer, AVR32_FLASHC_PAGE_SIZE)))
					{
					LCD_HD44780_write("ERR: P_VERIFY 2",2);
					usart_write_line(usart, "ERROR: ISP_STATE_PROGRAM_AND_VERIFY 2\r\n");
					isp_state = ISP_STATE_ERROR;
					break;
					}
					memset(&buffer[i], 0xFF, AVR32_FLASHC_PAGE_SIZE - i);
					flashc_clear_page_buffer();
					for (i = 0; i < AVR32_FLASHC_PAGE_SIZE / sizeof(*flash_ptr); i++)
					flash_ptr[i] = ((U64 *)buffer)[i];
					flashc_write_page(-1);
					if (memcmp((void *)flash_ptr, buffer, AVR32_FLASHC_PAGE_SIZE))
					{
					LCD_HD44780_write("ERR: P_VERIFY 3",2);
					usart_write_line(usart, "ERROR: ISP_STATE_PROGRAM_AND_VERIFY 3\r\n");
					isp_state = ISP_STATE_ERROR;
					break;
					}
					break;
	  
				case 0x01:
					file_close();                
	  
					// Clear the address AVR32_FLASHC_USER_PAGE_ADDRESS+ISP_FORCE_OFFSET ( 0x808001F8 )
					// where the key that activates de ISP programming is stored
					flash_ptr = (U64 *)AVR32_FLASHC_USER_PAGE;
					memcpy(buffer, (void *)flash_ptr, AVR32_FLASHC_PAGE_SIZE);
					*(U32 *)&buffer[ISP_FORCE_OFFSET] = 0xFFFFFFFF;
					flashc_clear_page_buffer();
					for (i = 0; i < AVR32_FLASHC_PAGE_SIZE / sizeof(*flash_ptr); i++)
					*flash_ptr++ = ((U64 *)buffer)[i];
					flashc_erase_user_page(FALSE);
					flashc_write_user_page();
     	  
					// enables the watchdog to reset de system
					Disable_global_interrupt();
					AVR32_WDT.ctrl = AVR32_WDT_CTRL_EN_MASK |
									(10 << AVR32_WDT_CTRL_PSEL_OFFSET) |
									(AVR32_WDT_KEY_VALUE << AVR32_WDT_CTRL_KEY_OFFSET);
					asm ("");
					AVR32_WDT.ctrl = AVR32_WDT_CTRL_EN_MASK |
									(10 << AVR32_WDT_CTRL_PSEL_OFFSET) |
									((~AVR32_WDT_KEY_VALUE << AVR32_WDT_CTRL_KEY_OFFSET) & AVR32_WDT_CTRL_KEY_MASK);
	  
					// wati until the system resets				   
					LCD_HD44780_write("READY RESET!",2);
					usart_write_line(usart, "READY RESET!\r\n");
					while (1){
					usart_write_line(usart, ".");  
					};
				case 0xFF:
					isp_state = ISP_STATE_ERROR;
					break;
			}//switch (file_eof())
			break;
	  }
}

// #endif  // USB_HOST_FEATURE == ENABLED


// Enable the serial port
void rs232_init(){
	
	  volatile avr32_usart_t  *usart;
	  usart_options_t USART_OPTIONS;
	  
	  USART_OPTIONS.baudrate=57600;
	  USART_OPTIONS.charlength=8;
	  USART_OPTIONS.paritytype=USART_NO_PARITY;
	  USART_OPTIONS.channelmode=USART_NORMAL_CHMODE;
	  USART_OPTIONS.stopbits=USART_1_STOPBIT;
	  
	  static const gpio_map_t USART_232_GPIO_MAP =
	  {
		  { serial_232_PORT_USART_RX_PIN, serial_232_PORT_USART_RX_FUNCTION },
		  { serial_232_PORT_USART_TX_PIN, serial_232_PORT_USART_TX_FUNCTION }
	  };
	  // Enable USART RXD & TXD pins.
	  gpio_enable_module( USART_232_GPIO_MAP, sizeof( USART_232_GPIO_MAP ) / sizeof( USART_232_GPIO_MAP[0] ) ); 
	  
	  usart = serial_232_PORT_USART;
	  
	  // Initialize USART in RS232 mode.
	  usart_init_rs232(usart, &USART_OPTIONS, FOSC0);
	  
	  // void usart_write_line(volatile avr32_usart_t *usart, const char *string)
	  usart_write_line(usart, "S-Bootloader\r\n");
	  
}//rs232_init



int main(void)
{
		
    sys_clk_gen_start();
	crc32_init(&crc32_table);
    // usb_task_init();

	// start the crystal oscillator 0 and switch the main clock to it.
	// pm_switch_to_osc0(&AVR32_PM, FOSC0, OSC0_STARTUP);

	gpio_local_init();

	// initialize use GIPIO pins
	// Enable the local bus interface for GPIO. gpio_local_init( ) must be called first!!
	gpio_local_enable_pin_output_driver(SCOREB_BUS_IO_01);
	gpio_local_enable_pin_output_driver(SCOREB_BUS_IO_02);
	gpio_local_enable_pin_output_driver(SCOREB_BUS_IO_03);
	gpio_local_enable_pin_output_driver(SCOREB_BUS_IO_04);
	gpio_local_enable_pin_output_driver(SCOREB_BUS_IO_05);
	gpio_local_enable_pin_output_driver(SCOREB_BUS_IO_06);
	gpio_local_enable_pin_output_driver(SCOREB_BUS_IO_07);
	gpio_local_enable_pin_output_driver(SCOREB_BUS_IO_08);
	gpio_local_enable_pin_output_driver(SCOREB_BUS_IO_09);
	gpio_local_enable_pin_output_driver(SCOREB_BUS_IO_10);
	gpio_local_enable_pin_output_driver(SCOREB_BUS_IO_11);
	gpio_local_enable_pin_output_driver(SCOREB_BUS_IO_12);
	gpio_local_enable_pin_output_driver(SCOREB_BUS_IO_13);
	gpio_local_enable_pin_output_driver(SCOREB_BUS_IO_14);
	gpio_local_enable_pin_output_driver(SCOREB_BUS_IO_15);
	gpio_local_enable_pin_output_driver(SCOREB_BUS_IO_16);
	gpio_local_enable_pin_output_driver(SCOREB_BUS_IO_17);
	gpio_local_enable_pin_output_driver(SCOREB_BUS_IO_18);
	gpio_local_enable_pin_output_driver(SCOREB_BUS_IO_19);
	gpio_local_enable_pin_output_driver(SCOREB_BUS_IO_20);
	gpio_local_enable_pin_output_driver(SCOREB_BUS_IO_21);
	gpio_local_enable_pin_output_driver(SCOREB_BUS_IO_22);
	gpio_local_enable_pin_output_driver(SCOREB_BUS_IO_23);
	gpio_local_enable_pin_output_driver(SCOREB_BUS_IO_24);
	gpio_local_enable_pin_output_driver(SCOREB_BUS_IO_25);
	gpio_local_enable_pin_output_driver(SCOREB_BUS_IO_26);
	gpio_local_enable_pin_output_driver(SCOREB_BUS_IO_27);
	gpio_local_enable_pin_output_driver(SCOREB_BUS_IO_28);
	gpio_local_enable_pin_output_driver(SCOREB_BUS_IO_29);
	gpio_local_enable_pin_output_driver(SCOREB_BUS_IO_30);
	gpio_local_enable_pin_output_driver(SCOREB_BUS_IO_31);
	gpio_local_enable_pin_output_driver(SCOREB_BUS_IO_32);
	gpio_local_enable_pin_output_driver(SCOREB_BUS_IO_33);
	gpio_local_enable_pin_output_driver(SCOREB_BUS_IO_34);
	gpio_local_enable_pin_output_driver(SCOREB_BUS_IO_35);

  
	rs232_init();

	FILE_SYS_SD_MMC_resources_init();

	LCD_HD44780_init();
	LCD_HD44780_clear();

	LCD_HD44780_write("Fwre Update",1);

 	while (TRUE)
	{
		  

// usb_task();
// #if USB_HOST_FEATURE == ENABLED
		  // host_mass_storage_task();
		  isp_task();
// #endif


	}

}
